@@ -3,23 +3,41 @@ Somewhat hacky solution to create conda lock files.
 """
 
 import atexit
+import datetime
 import json
 import logging
 import os
 import pathlib
+import re
+import shlex
 import shutil
 import subprocess
 import sys
 import tempfile
 
+from contextlib import contextmanager
+from functools import partial
 from itertools import chain
-from typing import Dict, List, MutableSequence, Optional, Sequence, Set, Tuple, Union
+from typing import (
+    Dict,
+    Iterator,
+    List,
+    MutableSequence,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Union,
+)
 
 import click
 import ensureconda
+import pkg_resources
 
 from click_default_group import DefaultGroup
 
+from conda_lock.common import read_file, read_json, write_file
+from conda_lock.errors import PlatformValidationError
 from conda_lock.src_parser import LockSpecification
 from conda_lock.src_parser.environment_yaml import parse_environment_file
 from conda_lock.src_parser.meta_yaml import parse_meta_yaml_file
@@ -28,6 +46,14 @@ from conda_lock.src_parser.pyproject_toml import parse_pyproject_toml
 
 PathLike = Union[str, pathlib.Path]
 
+# Captures basic auth credentials, if they exists, in the second capture group.
+AUTH_PATTERN = re.compile(r"^(https?:\/\/)(.*:.*@)?(.*)")
+
+# Captures the domain in the second group.
+DOMAIN_PATTERN = re.compile(r"^(https?:\/\/)?([^\/]+)(.*)")
+
+# Captures the platform in the first group.
+PLATFORM_PATTERN = re.compile(r"^# platform: (.*)$")
 
 if not (sys.version_info.major >= 3 and sys.version_info.minor >= 6):
     print("conda_lock needs to run under python >=3.6")
@@ -36,6 +62,49 @@ if not (sys.version_info.major >= 3 and sys.version_info.minor >= 6):
 
 CONDA_PKGS_DIRS = None
 DEFAULT_PLATFORMS = ["osx-64", "linux-64", "win-64"]
+DEFAULT_KINDS = ["explicit"]
+KIND_FILE_EXT = {
+    "explicit": "",
+    "env": ".yml",
+}
+KIND_USE_TEXT = {
+    "explicit": "conda create --name YOURENV --file {lockfile}",
+    "env": "conda env create --name YOURENV --file {lockfile}",
+}
+
+
+def _extract_platform(line: str) -> Optional[str]:
+    search = PLATFORM_PATTERN.search(line)
+    if search:
+        return search.group(1)
+    return None
+
+
+def extract_platform(lockfile: str) -> str:
+    for line in lockfile.strip().split("\n"):
+        platform = _extract_platform(line)
+        if platform:
+            return platform
+    raise RuntimeError("Cannot find platform in lockfile.")
+
+
+def _do_validate_platform(platform: str) -> Tuple[bool, str]:
+    from ensureconda.resolve import platform_subdir
+
+    determined_subdir = platform_subdir()
+    return platform == determined_subdir, platform
+
+
+def do_validate_platform(lockfile: str):
+    platform_lockfile = extract_platform(lockfile)
+    try:
+        success, platform_sys = _do_validate_platform(platform_lockfile)
+    except KeyError:
+        raise RuntimeError(f"Unknown platform type in lockfile '{platform_lockfile}'.")
+    if not success:
+        raise PlatformValidationError(
+            f"Platform in lockfile '{platform_lockfile}' is not compatible with system platform '{platform_sys}'."
+        )
 
 
 def conda_pkgs_dir():
@@ -73,6 +142,9 @@ def solve_specs_for_arch(
         "--dry-run",
         "--json",
     ]
+    conda_flags = os.environ.get("CONDA_FLAGS")
+    if conda_flags:
+        args.extend(shlex.split(conda_flags))
     if channels:
         args.append("--override-channels")
     for channel in channels:
@@ -123,18 +195,41 @@ def solve_specs_for_arch(
         sys.exit(1)
 
 
+def _process_stdout(stdout):
+    cache = set()
+    extracting_packages = False
+    leading_empty = True
+    for logline in stdout:
+        logline = logline.rstrip()
+        if logline:
+            leading_empty = False
+        if logline == "Downloading and Extracting Packages":
+            extracting_packages = True
+        if not logline and (extracting_packages or leading_empty):
+            continue
+        if "%" in logline:
+            logline = logline.split()[0]
+            if logline not in cache:
+                yield logline
+                cache.add(logline)
+        else:
+            yield logline
+
+
 def do_conda_install(conda: PathLike, prefix: str, name: str, file: str) -> None:
 
     if prefix and name:
         raise ValueError("Provide either prefix, or name, but not both.")
 
+    kind = "env" if file.endswith(".yml") else "explicit"
+
     args: MutableSequence[PathLike] = [
         str(conda),
+        *(["env"] if kind == "env" else []),
         "create",
         "--file",
         file,
-        "--yes",
-        "--json",
+        *([] if kind == "env" else ["--yes"]),
     ]
 
     if prefix:
@@ -143,38 +238,31 @@ def do_conda_install(conda: PathLike, prefix: str, name: str, file: str) -> None
     if name:
         args.append("--name")
         args.append(name)
+    conda_flags = os.environ.get("CONDA_FLAGS")
+    if conda_flags:
+        args.extend(shlex.split(conda_flags))
 
     logging.debug("$MAMBA_ROOT_PREFIX: %s", os.environ.get("MAMBA_ROOT_PREFIX"))
-    proc = subprocess.run(
+
+    with subprocess.Popen(
         args,
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
-        encoding="utf8",
-    )
-    logging.debug("install process: %s", proc)
+        bufsize=1,
+        universal_newlines=True,
+    ) as p:
+        if p.stdout:
+            for line in _process_stdout(p.stdout):
+                logging.info(line)
 
-    def print_proc(proc):
-        print(f"    Command: {proc.args}")
-        if proc.stdout:
-            print(f"    STDOUT:\n{proc.stdout}")
-        if proc.stderr:
-            print(f"    STDERR:\n{proc.stderr}")
-
-    try:
-        proc.check_returncode()
-    except subprocess.CalledProcessError:
-        try:
-            err_json = json.loads(proc.stdout)
-            message = err_json["message"]
-        except json.JSONDecodeError as e:
-            print(f"Failed to parse json, {e}")
-            message = ""
-
-        print(f"Could not perform conda install using {file} lock file into {prefix}")
-        if message:
-            print(message)
-        print_proc(proc)
+        if p.stderr:
+            for line in p.stderr:
+                logging.error(line.rstrip())
 
+    if p.returncode != 0:
+        print(
+            f"Could not perform conda install using {file} lock file into {name or prefix}"
+        )
         sys.exit(1)
 
 
@@ -244,9 +332,11 @@ def fn_to_dist_name(fn: str) -> str:
 def make_lock_files(
     conda: PathLike,
     platforms: List[str],
+    kinds: List[str],
     src_files: List[pathlib.Path],
     include_dev_dependencies: bool = True,
     channel_overrides: Optional[Sequence[str]] = None,
+    filename_template: Optional[str] = None,
 ):
     """Generate the lock files for the given platforms from the src file provided
 
@@ -262,10 +352,29 @@ def make_lock_files(
         For source types that separate out dev dependencies from regular ones,include those, default True
     channel_overrides :
         Forced list of channels to use.
+    filename_template :
+        Format for the lock file names. Must include {platform}.
 
     """
+    if filename_template:
+        if "{platform}" not in filename_template and len(platforms) > 1:
+            print(
+                "{platform} must be in filename template when locking"
+                f" more than one platform: {', '.join(platforms)}",
+                file=sys.stderr,
+            )
+            sys.exit(1)
+        for kind, file_ext in KIND_FILE_EXT.items():
+            if file_ext and filename_template.endswith(file_ext):
+                print(
+                    f"Filename template must not end with '{file_ext}', as this "
+                    f"is reserved for '{kind}' lock files, in which case it is "
+                    f"automatically added."
+                )
+                sys.exit(1)
+
     for plat in platforms:
-        print(f"generating lockfile for {plat}", file=sys.stderr)
+        print(f"Generating lockfile(s) for {plat}...", file=sys.stderr)
         lock_specs = parse_source_files(
             src_files=src_files,
             platform=plat,
@@ -278,28 +387,48 @@ def make_lock_files(
         else:
             channels = lock_spec.channels
 
-        lockfile_contents = create_lockfile_from_spec(
-            channels=channels, conda=conda, spec=lock_spec
-        )
-
-        def sanitize_lockfile_line(line):
-            line = line.strip()
-            if line == "":
-                return "#"
-            else:
-                return line
-
-        with open(f"conda-{lock_spec.platform}.lock", "w") as fo:
-            fo.write(
-                "\n".join(sanitize_lockfile_line(ln) for ln in lockfile_contents) + "\n"
+        for kind in kinds:
+            lockfile_contents = create_lockfile_from_spec(
+                channels=channels,
+                conda=conda,
+                spec=lock_spec,
+                kind=kind,
             )
 
-    print("To use the generated lock files create a new environment:", file=sys.stderr)
-    print("", file=sys.stderr)
-    print(
-        "     conda create --name YOURENV --file conda-linux-64.lock", file=sys.stderr
-    )
-    print("", file=sys.stderr)
+            if filename_template:
+                context = {
+                    "platform": lock_spec.platform,
+                    "dev-dependencies": str(include_dev_dependencies).lower(),
+                    "spec-hash": lock_spec.env_hash(),
+                    "version": pkg_resources.get_distribution("conda_lock").version,
+                    "timestamp": datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%SZ"),
+                }
+
+                filename = filename_template.format(**context)
+            else:
+                filename = f"conda-{lock_spec.platform}.lock"
+
+            filename += KIND_FILE_EXT[kind]
+            with open(filename, "w") as fo:
+                fo.write("\n".join(lockfile_contents) + "\n")
+
+            print(
+                f" - Install lock using {'(see warning below)' if kind == 'env' else ''}:",
+                KIND_USE_TEXT[kind].format(lockfile=filename),
+                file=sys.stderr,
+            )
+
+    if "env" in kinds:
+        print(
+            "\nWARNING: Using environment lock files (*.yml) does NOT guarantee "
+            "that generated environments will be identical over time, since the "
+            "dependency resolver is re-run every time and changes in repository "
+            "metadata or resolver logic may cause variation. Conversely, since "
+            "the resolver is run every time, the resulting packages ARE "
+            "guaranteed to be seen by conda as being in a consistent state. This "
+            "makes them useful when updating existing environments.",
+            file=sys.stderr,
+        )
 
 
 def is_micromamba(conda: PathLike) -> bool:
@@ -307,7 +436,11 @@ def is_micromamba(conda: PathLike) -> bool:
 
 
 def create_lockfile_from_spec(
-    *, channels: Sequence[str], conda: PathLike, spec: LockSpecification
+    *,
+    channels: Sequence[str],
+    conda: PathLike,
+    spec: LockSpecification,
+    kind: str,
 ) -> List[str]:
     dry_run_install = solve_specs_for_arch(
         conda=conda,
@@ -316,53 +449,80 @@ def create_lockfile_from_spec(
         specs=spec.specs,
     )
     logging.debug("dry_run_install:\n%s", dry_run_install)
+
     lockfile_contents = [
+        "# Generated by conda-lock.",
         f"# platform: {spec.platform}",
         f"# env_hash: {spec.env_hash()}\n",
-        "@EXPLICIT\n",
     ]
 
-    link_actions = dry_run_install["actions"]["LINK"]
-    for link in link_actions:
-        if is_micromamba(conda):
-            link["url_base"] = fn_to_dist_name(link["url"])
-            link["dist_name"] = fn_to_dist_name(link["fn"])
-        else:
-            link[
-                "url_base"
-            ] = f"{link['base_url']}/{link['platform']}/{link['dist_name']}"
-
-        link["url"] = f"{link['url_base']}.tar.bz2"
-        link["url_conda"] = f"{link['url_base']}.conda"
-    link_dists = {link["dist_name"] for link in link_actions}
-
-    fetch_actions = dry_run_install["actions"]["FETCH"]
-
-    fetch_by_dist_name = {fn_to_dist_name(pkg["fn"]): pkg for pkg in fetch_actions}
-
-    non_fetch_packages = link_dists - set(fetch_by_dist_name)
-    if len(non_fetch_packages) > 0:
-        for search_res in search_for_md5s(
-            conda=conda,
-            package_specs=[
-                x for x in link_actions if x["dist_name"] in non_fetch_packages
-            ],
-            platform=spec.platform,
-            channels=channels,
-        ):
-            dist_name = fn_to_dist_name(search_res["fn"])
-            fetch_by_dist_name[dist_name] = search_res
-
-    for pkg in link_actions:
-        dist_name = (
-            fn_to_dist_name(pkg["fn"]) if is_micromamba(conda) else pkg["dist_name"]
+    if kind == "env":
+        link_actions = dry_run_install["actions"]["LINK"]
+        lockfile_contents.extend(
+            [
+                "channels:",
+                *(f"  - {channel}" for channel in channels),
+                "dependencies:",
+                *(
+                    f'  - {pkg["name"]}={pkg["version"]}={pkg["build_string"]}'
+                    for pkg in link_actions
+                ),
+            ]
         )
-        url = fetch_by_dist_name[dist_name]["url"]
-        md5 = fetch_by_dist_name[dist_name]["md5"]
-        lockfile_contents.append(f"{url}#{md5}")
+    elif kind == "explicit":
+        lockfile_contents.append("@EXPLICIT\n")
+
+        link_actions = dry_run_install["actions"]["LINK"]
+        for link in link_actions:
+            if is_micromamba(conda):
+                link["url_base"] = fn_to_dist_name(link["url"])
+                link["dist_name"] = fn_to_dist_name(link["fn"])
+            else:
+                link[
+                    "url_base"
+                ] = f"{link['base_url']}/{link['platform']}/{link['dist_name']}"
+
+            link["url"] = f"{link['url_base']}.tar.bz2"
+            link["url_conda"] = f"{link['url_base']}.conda"
+        link_dists = {link["dist_name"] for link in link_actions}
+
+        fetch_actions = dry_run_install["actions"]["FETCH"]
+
+        fetch_by_dist_name = {fn_to_dist_name(pkg["fn"]): pkg for pkg in fetch_actions}
+
+        non_fetch_packages = link_dists - set(fetch_by_dist_name)
+        if len(non_fetch_packages) > 0:
+            for search_res in search_for_md5s(
+                conda=conda,
+                package_specs=[
+                    x for x in link_actions if x["dist_name"] in non_fetch_packages
+                ],
+                platform=spec.platform,
+                channels=channels,
+            ):
+                dist_name = fn_to_dist_name(search_res["fn"])
+                fetch_by_dist_name[dist_name] = search_res
+
+        for pkg in link_actions:
+            dist_name = (
+                fn_to_dist_name(pkg["fn"]) if is_micromamba(conda) else pkg["dist_name"]
+            )
+            url = fetch_by_dist_name[dist_name]["url"]
+            md5 = fetch_by_dist_name[dist_name]["md5"]
+            lockfile_contents.append(f"{url}#{md5}")
+
+        def sanitize_lockfile_line(line):
+            line = line.strip()
+            if line == "":
+                return "#"
+            else:
+                return line
+
+        lockfile_contents = [sanitize_lockfile_line(line) for line in lockfile_contents]
+    else:
+        raise ValueError(f"Unrecognised lock kind {kind}.")
 
     logging.debug("lockfile_contents:\n%s\n", lockfile_contents)
-
     return lockfile_contents
 
 
@@ -467,6 +627,61 @@ def determine_conda_executable(
     raise RuntimeError("Could not find conda (or compatible) executable")
 
 
+def _add_auth_to_line(line: str, auth: Dict[str, str]):
+    search = DOMAIN_PATTERN.search(line)
+    if search and search.group(2) in auth:
+        return f"{search.group(1)}{auth[search.group(2)]}@{search.group(2)}{search.group(3)}"
+    return line
+
+
+def _add_auth_to_lockfile(lockfile: str, auth: Dict[str, str]) -> str:
+    lockfile_with_auth = "\n".join(
+        _add_auth_to_line(line, auth) if line[0] not in ("#", "@") else line
+        for line in lockfile.strip().split("\n")
+    )
+    if lockfile.endswith("\n"):
+        return lockfile_with_auth + "\n"
+    return lockfile_with_auth
+
+
+@contextmanager
+def _add_auth(lockfile: str, auth: Dict[str, str]) -> Iterator[str]:
+    with tempfile.NamedTemporaryFile() as tf:
+        lockfile_with_auth = _add_auth_to_lockfile(lockfile, auth)
+        write_file(lockfile_with_auth, tf.name)
+        yield tf.name
+
+
+def _strip_auth_from_line(line: str) -> str:
+    return AUTH_PATTERN.sub(r"\1\3", line)
+
+
+def _extract_domain(line: str) -> str:
+    return DOMAIN_PATTERN.sub(r"\2", line)
+
+
+def _strip_auth_from_lockfile(lockfile: str) -> str:
+    lockfile_lines = lockfile.strip().split("\n")
+    stripped_lockfile_lines = tuple(
+        _strip_auth_from_line(line) if line[0] not in ("#", "@") else line
+        for line in lockfile_lines
+    )
+    stripped_domains = sorted(
+        {
+            _extract_domain(stripped_line)
+            for line, stripped_line in zip(lockfile_lines, stripped_lockfile_lines)
+            if line != stripped_line
+        }
+    )
+    stripped_lockfile = "\n".join(stripped_lockfile_lines)
+    if lockfile.endswith("\n"):
+        stripped_lockfile += "\n"
+    if stripped_domains:
+        stripped_domains_doc = "\n".join(f"# - {domain}" for domain in stripped_domains)
+        return f"# The following domains require authentication:\n{stripped_domains_doc}\n{stripped_lockfile}"
+    return stripped_lockfile
+
+
 def run_lock(
     environment_files: List[pathlib.Path],
     conda_exe: Optional[str],
@@ -475,6 +690,8 @@ def run_lock(
     micromamba: bool = False,
     include_dev_dependencies: bool = True,
     channel_overrides: Optional[Sequence[str]] = None,
+    filename_template: Optional[str] = None,
+    kinds: Optional[List[str]] = None,
 ) -> None:
     _conda_exe = determine_conda_executable(
         conda_exe, mamba=mamba, micromamba=micromamba
@@ -485,6 +702,8 @@ def run_lock(
         platforms=platforms or DEFAULT_PLATFORMS,
         include_dev_dependencies=include_dev_dependencies,
         channel_overrides=channel_overrides,
+        filename_template=filename_template,
+        kinds=kinds or DEFAULT_KINDS,
     )
 
 
@@ -534,6 +753,31 @@ def main():
     multiple=True,
     help="path to a conda environment specification(s)",
 )
+@click.option(
+    "-k",
+    "--kind",
+    default=["explicit"],
+    type=str,
+    multiple=True,
+    help="Kind of lock file(s) to generate [should be one of 'explicit' or 'env'].",
+)
+@click.option(
+    "--filename-template",
+    default="conda-{platform}.lock",
+    help="Template for the lock file names. Filename must include {platform} token, and must not end in '.yml'. For a full list and description of available tokens, see the command help text.",
+)
+@click.option(
+    "--strip-auth",
+    is_flag=True,
+    default=False,
+    help="Strip the basic auth credentials from the lockfile.",
+)
+@click.option(
+    "--log-level",
+    help="Log level.",
+    default="INFO",
+    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
+)
 # @click.option(
 #     "-m",
 #     "--mode",
@@ -545,11 +789,34 @@ def main():
 #             existing condarc configurations.""",
 # )
 def lock(
-    conda, mamba, micromamba, platform, channel_overrides, dev_dependencies, files
+    conda,
+    mamba,
+    micromamba,
+    platform,
+    channel_overrides,
+    dev_dependencies,
+    files,
+    kind,
+    filename_template,
+    strip_auth,
+    log_level,
 ):
-    """Generate fully reproducible lock files for conda environments."""
+    """Generate fully reproducible lock files for conda environments.
+
+    By default, the lock files are written to conda-{platform}.lock. These filenames can be customized using the
+    --filename-template argument. The following tokens are available:
+
+    \b
+        platform: The platform this lock file was generated for (conda subdir).
+        dev-dependencies: Whether or not dev dependencies are included in this lock file.
+        spec-hash: A sha256 hash of the lock file spec.
+        version: The version of conda-lock used to generate this lock file.
+        timestamp: The approximate timestamp of the output file in ISO8601 basic format.
+    """
+    logging.basicConfig(level=log_level)
     files = [pathlib.Path(file) for file in files]
-    run_lock(
+    lock_func = partial(
+        run_lock,
         environment_files=files,
         conda_exe=conda,
         platforms=platform,
@@ -557,7 +824,19 @@ def lock(
         micromamba=micromamba,
         include_dev_dependencies=dev_dependencies,
         channel_overrides=channel_overrides,
+        kinds=kind,
     )
+    if strip_auth:
+        with tempfile.TemporaryDirectory() as tempdir:
+            filename_template_temp = f"{tempdir}/{filename_template.split('/')[-1]}"
+            lock_func(filename_template=filename_template_temp)
+            filename_template_dir = "/".join(filename_template.split("/")[:-1])
+            for file in os.listdir(tempdir):
+                lockfile = read_file(os.path.join(tempdir, file))
+                lockfile = _strip_auth_from_lockfile(lockfile)
+                write_file(lockfile, os.path.join(filename_template_dir, file))
+    else:
+        lock_func(filename_template=filename_template)
 
 
 @main.command("install")
@@ -574,11 +853,56 @@ def lock(
 )
 @click.option("-p", "--prefix", help="Full path to environment location (i.e. prefix).")
 @click.option("-n", "--name", help="Name of environment.")
+@click.option(
+    "--auth",
+    help="The auth file provided as string. Has precedence over `--auth-file`.",
+    default="",
+)
+@click.option("--auth-file", help="Path to the authentication file.", default="")
+@click.option(
+    "--validate-platform",
+    is_flag=True,
+    default=True,
+    help="Whether the platform compatibility between your lockfile and the host system should be validated.",
+)
+@click.option(
+    "--log-level",
+    help="Log level.",
+    default="INFO",
+    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
+)
 @click.argument("lock-file")
-def install(conda, mamba, micromamba, prefix, name, lock_file):
+def install(
+    conda,
+    mamba,
+    micromamba,
+    prefix,
+    name,
+    lock_file,
+    auth,
+    auth_file,
+    validate_platform,
+    log_level,
+):
     """Perform a conda install"""
+    logging.basicConfig(level=log_level)
+    auth = json.loads(auth) if auth else read_json(auth_file) if auth_file else None
     _conda_exe = determine_conda_executable(conda, mamba=mamba, micromamba=micromamba)
-    do_conda_install(conda=_conda_exe, prefix=prefix, name=name, file=lock_file)
+    install_func = partial(do_conda_install, conda=_conda_exe, prefix=prefix, name=name)
+    if validate_platform:
+        lockfile = read_file(lock_file)
+        try:
+            do_validate_platform(lockfile)
+        except PlatformValidationError as error:
+            raise PlatformValidationError(
+                error.args[0] + " Disable validation with `--validate-platform=False`."
+            )
+    if auth:
+        lockfile = read_file(lock_file)
+        with _add_auth(lockfile, auth) as lockfile_with_auth:
+            install_func(file=lockfile_with_auth)
+    else:
+        install_func(file=lock_file)
 
 
 if __name__ == "__main__":
